import * as PIXI from "pixi.js";
import "pixi-plugin-bump";
const { Application, Sprite, Container, TextStyle, Text } = PIXI;
const { TextureCache } = PIXI.utils;
import { loaderRes } from "./util";
import createGroundData from "./groudData";

const GAME_RESULT_STATE = {
    ON: "游戏开始",
    OVER: "游戏失败",
    SUCC: "通关完成"
};
const ROLESTATE = {
    READY: "READY",
    RUN: "RUN",
    JUMP: "JUMP",
    DOWN: "DOWN"
};

class Game {
    constructor(wrapper) {
        this.stageWidth = 750;
        this.stageHeight = 1334;

        this.gameState = GAME_RESULT_STATE.ON;
        this.wrapper = wrapper;
        this.roleJump = false; //跳起状态
        this.fallingDown = false; //下坠状态
        this.decibel = 0; //分贝总和
        //ui
        this.uiGameTopTip = null; //游戏顶部提示
        this.uiCountdown = null; //游戏开始倒计时
        this.uiFireworksContanier = null; //烟花容器
        this.fireworkOpts = {
            green: {
                delay: 0,
                pos: {
                    x: 588,
                    y: 230
                }
            },
            orange: {
                delay: 300,
                pos: {
                    x: 290,
                    y: 375
                }
            },
            purple: {
                delay: 500,
                pos: {
                    x: 400,
                    y: 200
                }
            },
            red: {
                delay: 600,
                pos: {
                    x: 450,
                    y: 370
                }
            }
        };
        this.init();
        this.initProxy();
    }

    //底部小游戏
    init() {
        this.app = new Application({
            width: this.stageWidth,
            height: this.stageHeight,
            transparent: true
        });
        this.wrapper.appendChild(this.app.view);
        this.bump = new PIXI.extras.Bump();
        loaderRes(["/scene.json"], this.loadedSource, this);
    }
    initProxy() {
        Object.defineProperty(this, "roleState", {
            configurable: true,
            enumerable: true,
            get: () => {
                return this._roleState;
            },
            set: value => {
                if (this._roleState === value) return;
                this._roleState = value;
                this.clearRoleStateEffect();

                switch (value) {
                    case ROLESTATE.READY:
                        this.roleReadyStep = 0;
                        this.roleReadyLoop();
                        break;
                    case ROLESTATE.RUN:
                        this.roleRunStep = 0;
                        this.roleRunLoop();
                        break;
                    case ROLESTATE.JUMP:
                        this.role.texture = TextureCache[`p_jump_0.png`];
                        break;
                    case ROLESTATE.DOWN:
                        this.role.texture = TextureCache[`p_jump_1.png`];
                        break;
                }
            }
        });
    }
    clearRoleStateEffect() {
        clearTimeout(this.roleReadyTimer);
        clearTimeout(this.roleRunTimer);
    }
    roleReadyLoop() {
        this.roleReadyTimer = setTimeout(() => {
            this.role.texture =
                TextureCache[`p_stand_${this.roleReadyStep % 2}.png`];
            this.roleReadyStep++;
            this.roleReadyLoop();
        }, 500);
    }

    roleRunLoop() {
        this.roleRunTimer = setTimeout(() => {
            this.role.texture =
                TextureCache[`p_run_${this.roleRunStep % 2}.png`];
            this.roleRunStep++;
            this.roleRunLoop();
        }, 100);
    }

    loadedSource() {
        this.initGround();
        this.initRole();
        this.initGameReady();
        this.gameState = GAME_RESULT_STATE.ON;
        this.readyStart();
        this.initFireworks();
    }
    readyStart() {
        this.uiCountdownTimes = 3;
        this.uiCountdown.texture = TextureCache[`word_3.png`];
        this.uiCountdown.visible = true;
        this.uiCountdownTimer = setTimeout(this.countDownFn.bind(this), 1000);
    }
    countDownFn() {
        this.uiCountdownTimes--;
        if (this.uiCountdownTimes === 0) {
            this.uiCountdown.texture = TextureCache[`word_start.png`];
            this.gameStart();
        } else if (this.uiCountdownTimes < 0) {
            clearTimeout(this.uiCountdownTimer);
            this.uiCountdown.visible = false;
        } else {
            this.uiCountdown.texture =
                TextureCache[`word_${this.uiCountdownTimes}.png`];
        }

        this.uiCountdownTimer = setTimeout(this.countDownFn.bind(this), 1000);
    }
    gameStart() {
        this.groundWalk();
        this.roleState = ROLESTATE.RUN;
    }
    //创建角色
    initRole() {
        this.role = Sprite.fromFrame("people.png");
        this.role.x = this.stageWidth / 2;
        this.role.y = this.stageHeight - this.groundData[0].height + 24; //初始化位置
        this.role._vy = 4;
        this.role.anchor.x = 0.5;
        this.role.anchor.y = 1;
        this.app.stage.addChild(this.role);
        this.app.ticker.add(this.jumpLoop, this);
        this.roleState = ROLESTATE.READY;
    }
    // 创建地面容器
    initGround() {
        this.groundArr = [];
        this.totalWidth = 0;
        this.groundData = createGroundData();
        this.groundContainer = new Container();
        this.groundContainer._vx = 2.5;
        this.app.stage.addChild(this.groundContainer);
        this.groundData.forEach(this.createGround.bind(this));
        this.totalWidth = this.groundContainer.width;
    }
    //生成地面
    createGround(groundData, index, arr) {
        let { height, blockNum, gap } = groundData;
        this.totalWidth += gap;
        const groundUnitContainer = new Container();
        groundData.x = groundUnitContainer.x = this.totalWidth;
        groundData.y = groundUnitContainer.y = this.stageHeight - height + 24; //草地高度除外
        groundData.width = blockNum * 100;
        groundUnitContainer.cacheAsBitmap = true;
        let totalX = 0;
        for (let i = 0; i < blockNum; i++) {
            const ground = Sprite.fromFrame(
                `bg_${i === 0 ? "l" : i === blockNum - 1 ? "r" : "m"}.png`
            );
            ground.x = totalX;
            ground.y = -24; //负值，刚好贴着地面。方便碰撞检测
            groundUnitContainer.addChild(ground);
            totalX += ground.width;
        }
        this.totalWidth += groundData.width;
        this.groundArr.push(groundUnitContainer);
        this.groundContainer.addChild(groundUnitContainer);

        //最后一个，需要添加通关门UI
        if (index === arr.length - 1) {
            let uiCastle = Sprite.from("castle.png");
            uiCastle.anchor.x = 0.7;
            uiCastle.anchor.y = 1;
            uiCastle.x = totalX;
            groundUnitContainer.addChild(uiCastle);
        }
    }

    walkLoop() {
        this.groundContainer.x -= this.groundContainer._vx;
        let rolePosX = Math.abs(this.groundContainer.x - 375); //人物所在位置
        let rolePosY = this.role.y;

        if (rolePosX >= this.totalWidth - 375) {
            console.log("完成游戏");
            this.gameSucc();
            return;
        } else if (rolePosY >= this.stageHeight) {
            console.log("掉坑里挂了");
            this.gameOver();
            return;
        }

        let index = this.groundData.findIndex(
            ground =>
                rolePosX >= ground.x && rolePosX <= ground.x + ground.width
        );
        if (index >= 0) {
            /**
             * 人物下方的方块下标
             * 判断是否和方块有碰撞
             */
            let ground = this.groundArr[index];
            if (ground.y <= rolePosY) {
                let collision = this.bump.hit(
                    this.role,
                    ground,
                    true,
                    true,
                    true
                );
                console.log("collision", collision);
                if (collision) {
                    switch (collision) {
                        case "bottom":
                            console.log("站上了");
                            this.standGround();
                            break;
                        default:
                            console.log("撞上了");
                            this.gameOver();
                            break;
                    }
                }
            }
        } else {
            /**
             * 掉坑里了
             * 如果不是跳起状态，默认下坠
             *  */
            console.log("在坑里的范围内");
            if (!this.roleJump && !this.fallingDown) {
                this.fallDown();
            }
        }
    }

    standGround() {
        this.roleState = ROLESTATE.RUN;
        this.app.ticker.remove(this.downLoop, this);
        this.roleJump = false;
        this.fallingDown = false;
    }

    //开始倒计时
    initGameReady() {
        this.uiGameTopTip = Sprite.from("word_tips.png");
        this.uiGameTopTip.anchor.x = 0.5;
        this.uiGameTopTip.x = 375;
        this.uiGameTopTip.y = 20;
        this.app.stage.addChild(this.uiGameTopTip);

        this.uiCountdown = Sprite.from("word_3.png");
        this.uiCountdown.anchor.x = 0.5;
        this.uiCountdown.anchor.y = 0.5;
        this.uiCountdown.x = 375;
        this.uiCountdown.y = 400;
        this.app.stage.addChild(this.uiCountdown);
    }

    gameOver() {
        this.roleJump = false;
        this.fallingDown = false;
        this.app.ticker.remove(this.fallDownLoop, this);
        this.gameState = GAME_RESULT_STATE.OVER;
        this.app.ticker.stop();
    }

    gameSucc() {
        this.groundStop();
        this.roleWalk();
    }

    restartGame() {
        console.log("restartGame");
        this.gameState = GAME_RESULT_STATE.ON;
        this.groundContainer.x = 0;
        this.role.x = 375;
        this.role.y = this.groundData[0].y;
        this.gameResultContainer.visible = false;
        this.decibel = 0;
        this.roleState = ROLESTATE.RUN;
        this.app.ticker.start();
    }
    roleWalk(){
        this.app.ticker.add(this.roleWalkLoop, this);
    }
    roleStop(){
        this.app.ticker.remove(this.roleWalkLoop, this);
    }
    roleWalkLoop() {
        this.role.x += 2.5;
        if (this.role.x >= 750 + this.role.width / 2) {
            console.log("人物通关");
            this.roleStop();
            this.startFireworks();
        }
    }

    groundWalk() {
        this.app.ticker.add(this.walkLoop, this);
    }

    groundStop() {
        this.app.ticker.remove(this.walkLoop, this);
    }
    // 跳起，有一个最高值
    jump(decibel) {
        if (this.gameState === GAME_RESULT_STATE.OVER) return;
        if (this.decibel >= 80) return;
        this.decibel += decibel;
        this.roleJump = true;
        this.fallingDown = false;
        this.app.ticker.remove(this.fallDownLoop, this);
        this.app.ticker.remove(this.downLoop, this);
    }
    jumpLoop() {
        if (this.decibel > 0) {
            let distance = Math.min(this.role._vy, this.decibel);
            this.role.y -= distance;
            this.decibel -= distance;
            console.log("distance,decibel", distance, this.decibel);
            this.roleState = ROLESTATE.JUMP;
            if (this.decibel === 0) {
                this.app.ticker.add(this.downLoop, this);
            }
        }
    }

    downLoop() {
        this.roleJump = true;
        this.roleState = ROLESTATE.DOWN;
        this.role.y += this.role._vy;
    }

    fallDownLoop() {
        this.role.y += this.role._vy;
    }

    fallDown() {
        this.fallingDown = true;
        this.roleState = ROLESTATE.DOWN;
        this.app.ticker.add(this.fallDownLoop, this);
    }

    initFireworks() {
        this.uiFireworksContanier = new Container();
        Object.keys(this.fireworkOpts).forEach(key => {
            let firework = this.fireworkOpts[key];
            firework.frameFirework = 0; //下标
            firework.timer = null; //计时器

            let uiFirework = Sprite.from(
                `fireworks_${key}_${firework.frameFirework}.png`
            ); //ui
            uiFirework.anchor.x = 0.5;
            uiFirework.anchor.y = 0.5;
            uiFirework.x = firework.pos.x;
            uiFirework.y = firework.pos.y;

            firework.uiFirework = uiFirework;

            this.uiFireworksContanier.addChild(uiFirework);
        });
        this.uiFireworksContanier.visible = false;
        this.app.stage.addChild(this.uiFireworksContanier);
    }

    startFireworks() {
        this.uiFireworksContanier.visible = true;

        Object.keys(this.fireworkOpts).forEach(key => {
            let firework = this.fireworkOpts[key];
            firework.frameFirework = 0;
            clearInterval(firework.timer);
            setTimeout(() => {
                firework.timer = setInterval(() => {
                    firework.uiFirework.visible = (firework.frameFirework <= 6);
                    if (firework.frameFirework <= 6) {
                        firework.uiFirework.texture =
                            TextureCache[
                                `fireworks_${key}_${firework.frameFirework}.png`
                            ];
                    }
                    firework.frameFirework = (firework.frameFirework + 1) % 12;
                }, 100);
            }, firework.delay);
        });
    }
}

export default Game;
